@@ -193,11 +193,30 @@ func packWalkFn(root, src, dst string, tarW *tar.Writer, meta *Meta, dereference
 			header.Size = info.Size()
 
 		case fm&os.ModeSymlink != 0:
-			target, err := filepath.EvalSymlinks(path)
+			// First read the symlink file to find the destination.
+			target, err := os.Readlink(path)
 			if err != nil {
 				return fmt.Errorf("failed to get symbolic link destination for %q: %w", path, err)
 			}
 
+			// Try to make absolute paths relative.
+			if filepath.IsAbs(target) {
+				absPath, err := filepath.Abs(path)
+				if err != nil {
+					return fmt.Errorf("failed to get absolute path for %q: %w", path, err)
+				}
+				absDir := filepath.Dir(absPath)
+
+				rel, err := filepath.Rel(absDir, target)
+				if err != nil {
+					return fmt.Errorf("failed to find relative path for %q: %w", target, err)
+				}
+				target = rel
+			}
+
+			// Get the path to the target relative to path.
+			target = filepath.Join(filepath.Dir(path), target)
+
 			// If the target is within the current source, we
 			// create the symlink using a relative path.
 			if strings.HasPrefix(target, src) {
@@ -215,9 +234,13 @@ func packWalkFn(root, src, dst string, tarW *tar.Writer, meta *Meta, dereference
 			}
 
 			if !dereference {
-				// Return early as the symlink has a target outside of the
-				// src directory and we don't want to dereference symlinks.
-				return nil
+				// Symlinks with targets outside of src are prohibited.
+				return &IllegalSlugError{
+					Err: fmt.Errorf(
+						"invalid symlink (%q -> %q) has target outside of %q",
+						path, target, src,
+					),
+				}
 			}
 
 			// Get the file info for the target.
@@ -386,7 +409,7 @@ func Unpack(r io.Reader, dst string) error {
 		}
 
 		// Only unpack regular files from this point on.
-		if header.Typeflag == tar.TypeDir {
+		if header.Typeflag == tar.TypeDir || header.Typeflag == tar.TypeXGlobalHeader || header.Typeflag == tar.TypeXHeader {
 			continue
 		} else if header.Typeflag != tar.TypeReg && header.Typeflag != tar.TypeRegA {
 			return fmt.Errorf("failed creating %q: unsupported type %c", path,
